@@ -26,8 +26,6 @@
  *   that compiling, linking, and/or using OpenSSL is allowed.
  *   
  *   
- *   $Rev: 162 $
- *   $Date: 2013-03-08 14:56:32 +0100 (Fri, 08 Mar 2013) $
  *   
  *   
 */
@@ -141,7 +139,9 @@ OphcrackGUI::OphcrackGUI(ophcrack_t *crack) : QMainWindow() {
   tableModel->insertRoot(lmflash);
   tableModel->insertRoot(nteightxl);
   tableModel->insertRoot(ntspecialxl);
-  tableModel->insertRoot(probaTable);
+  tableModel->insertRoot(ntprobafree);
+  tableModel->insertRoot(ntproba10g);
+  tableModel->insertRoot(ntproba60g);
 
   // Associate a menu to the load button.
 
@@ -1349,28 +1349,6 @@ void OphcrackGUI::saveConfig(void) {
 
   if (file == 0) return;
 
-  // Save the table paths. Use a map in order to be sure that each
-  // path is saved only once.
-
-  QMap<QString, int> paths;
-  list_nd_t *nd;
-
-  for (nd = arg->table_path->head; nd != 0; nd = nd->next) {
-    QString path = QString((char*)nd->data);
-    
-    if (path[0] != 0)
-      paths.insert(path, 1);
-  }
-
-  QList<QString> keys = paths.uniqueKeys();
-
-  while (!keys.isEmpty()) {
-    QString str = keys.takeFirst();
-    const char *path = str.toLatin1().constData();
-
-    fprintf(file, "table_path %s\n", path);
-  }
-
   // Save the order and state (enabled or not) of the tables.
 
   int nrows = tableModel->rowCount();
