@@ -26,8 +26,6 @@
  *   that compiling, linking, and/or using OpenSSL is allowed.
  *   
  *   
- *   $Rev: 162 $
- *   $Date: 2013-03-08 14:56:32 +0100 (Fri, 08 Mar 2013) $
  *   
  *   
 */
@@ -56,7 +54,7 @@
 #include "lmflash.h"
 #include "nteightxl.h"
 #include "ntspecialxl.h"
-#include "probaTable.h"
+#include "ntproba.h"
 #include "message.h"
 /*-------------------------------------------------------------------------*/
 ophcrack_t *ophcrack_alloc(int nthreads, arg_t *arg) {
@@ -73,8 +71,10 @@ ophcrack_t *ophcrack_alloc(int nthreads, arg_t *arg) {
   crack->stat = ophstat_alloc();
   crack->maxtid = 0;
   crack->maxhid = 0;
+  crack->maxtkind = unknown;
   crack->npwds_total = 0;
   crack->npwds_found = 0;
+  crack->find_freeram = find_freeram;
   crack->freeram = 0;
   crack->arg = arg;
   
@@ -82,13 +82,21 @@ ophcrack_t *ophcrack_alloc(int nthreads, arg_t *arg) {
 }
 /*-------------------------------------------------------------------------*/
 void ophcrack_free(ophcrack_t *crack) {
-  /* ... */
+  scheduler_free(crack->sched);
+
+  list_free(crack->tables);
+  list_free(crack->hashes);
+  list_free(crack->active);
+  list_free(crack->enabled);
+  list_free(crack->remaining);
+
+  ophstat_free(crack->stat);
 
   free(crack);
 }
 /*-------------------------------------------------------------------------*/
 void ophcrack_start(ophcrack_t *crack) {
-  crack->freeram = find_freeram();
+  crack->freeram = crack->find_freeram();
   scheduler_start(crack->sched, ophcrack_run, crack);
 }
 /*-------------------------------------------------------------------------*/
@@ -162,8 +170,16 @@ int ophcrack_setup_table(table_t *tbl) {
     ret = ntspecialxl_setup(tbl);
     break;
 
-  case probaTable:
-    ret = probaTable_setup(tbl);
+  case ntprobafree:
+    ret = ntproba_setup(tbl);
+    break;
+
+  case ntproba10g:
+    ret = ntproba_setup(tbl);
+    break;
+
+  case ntproba60g:
+    ret = ntproba_setup(tbl);
     break;
 
   case unknown:
@@ -178,6 +194,8 @@ void ophcrack_add_table(ophcrack_t *crack, table_t *tbl) {
   scheduler_t *sched = crack->sched;
   pthread_mutex_t *mutex = sched->mutex;
 
+  crack->maxtkind = MY_MAX(crack->maxtkind, tbl->kind+1);
+
   tbl->id   = crack->maxtid++;
   tbl->cmin = tbl->ncols;
   tbl->cmax = tbl->ncols;
@@ -543,6 +561,8 @@ ophtask_t *ophcrack_find(ophcrack_t *crack, ophtask_t *task, int id) {
       fprintf(arg->lfile, "at column %d ... ", cmax-i);
     }
 
+    assert(tbl->param != NULL);
+
     tosearch[i]->col = cmax-i;
     tbl->find(hsh, tbl, tosearch[i]);
 
@@ -562,7 +582,7 @@ ophtask_t *ophcrack_find(ophcrack_t *crack, ophtask_t *task, int id) {
   /* If the .index file has been preloaded, then we proceed with the
      corresponding lookup task. */
 
-  if (tbl->preload > 0)
+  if (tbl->idxmem)
     task = ophcrack_lookup_idx(crack, task, id);
 
   /* Otherwise, we add the lookup task to the scheduler. */
@@ -600,10 +620,9 @@ ophtask_t *ophcrack_lookup_idx(ophcrack_t *crack, ophtask_t *task, int id) {
   /* If we access the disk, then it is better to query the .index file
      by increasing order of the index. */
 
-  int preload = tbl->preload;
   ophel_t **tosearch = work->tosearch;
 
-  if (preload < 1)
+  if (!tbl->idxmem)
     qsort(tosearch, size, sizeof(ophel_t*), ophcrack_pwd_cmp);
   
   /* Find the lookup range of each prefix. */
@@ -635,6 +654,7 @@ ophtask_t *ophcrack_lookup_idx(ophcrack_t *crack, ophtask_t *task, int id) {
       fprintf(arg->lfile, "and postfix = %u ... ", (unsigned int)postfix);
     }
 
+    assert(tbl->param != NULL);
     tbl->lookup_idx(hsh, tbl, tosearch[i]);
 
     if (arg->debug) {
@@ -646,16 +666,20 @@ ophtask_t *ophcrack_lookup_idx(ophcrack_t *crack, ophtask_t *task, int id) {
     }
   }
 
-  /* If no files have been preloaded or if the other files have been
-     preloaded, then we continue. */
+  /* If we have been searching in memory (on disk) and we can continue
+     do so, we go ahead. */
 
-  if (preload != 1)
+  int both_in_mem  =  tbl->idxmem &&  tbl->endmem;
+  int both_on_disk = !tbl->idxmem && !tbl->endmem;
+
+  if (both_in_mem || both_on_disk)
     task = ophcrack_lookup_end(crack, task, id);
 
-  /* Otherwise, we add a disk task to the scheduler. */
+  /* Otherwise, we add a disk (memory) task to the scheduler. */
 
   else {
-    scheduler_add(sched, task, disk);
+    prty_t priority = tbl->endmem ? low : disk;
+    scheduler_add(sched, task, priority);
     task = 0;
   }
 
@@ -718,6 +742,7 @@ ophtask_t *ophcrack_lookup_end(ophcrack_t *crack, ophtask_t *task, int id) {
       fprintf(arg->lfile, "low = %llu and high = %llu ... ", (long long unsigned)low, (long long unsigned)high);
     }
 
+    assert(tbl->param != NULL);
     int ret = tbl->lookup_end(hsh, tbl, tosearch[i]);
 
     /* Accumulate the statistics before freeing the element. */
@@ -749,16 +774,20 @@ ophtask_t *ophcrack_lookup_end(ophcrack_t *crack, ophtask_t *task, int id) {
   else
     return task;
 
-  /* If less then two files have been preloaded or if the .start files
-     has been preloaded, then we continue. */
+  /* If we have been searching in memory (on disk) and we can continue
+     do so, we go ahead. */
 
-  if (tbl->preload != 2)
+  int both_in_mem  =  tbl->endmem &&  tbl->srtmem;
+  int both_on_disk = !tbl->endmem && !tbl->srtmem;
+
+  if (both_in_mem || both_on_disk)
     task = ophcrack_lookup_srt(crack, task, id);
 
-  /* Otherwise, we send a disk task to the scheduler. */
+  /* Otherwise, we add a disk (memory) task to the scheduler. */
 
   else {
-    scheduler_add(sched, task, disk);
+    prty_t priority = tbl->srtmem ? low : disk;
+    scheduler_add(sched, task, priority);
     task = 0;
   }
 
@@ -824,6 +853,7 @@ ophtask_t *ophcrack_lookup_srt(ophcrack_t *crack, ophtask_t *task, int id) {
       fprintf(arg->lfile, "and offset = %llu ... ", (long long unsigned int)offset);
     }
     
+    assert(tbl->param != NULL);
     tbl->lookup_srt(hsh, tbl, tosearch[i]);
 
     if (arg->debug) {
@@ -901,6 +931,8 @@ ophtask_t *ophcrack_check(ophcrack_t *crack, ophtask_t *task, int id) {
 
     /* Check if the correct password has been found. */
 
+    assert(tbl->param != NULL);
+
     int col = tosearch[i]->col;
     char *pwd = tosearch[i]->pwd;
     int found = tbl->check(hsh, tbl, tosearch[i]);
@@ -1142,7 +1174,9 @@ ophtask_t *ophcrack_bforce_nt(ophcrack_t *crack, ophtask_t *task, int id) {
       
       ntextended_mkhash(pwd, hash);
 
-      bloom_byte = (* (int*)hash) & 0xfffff;
+      int *bbp = (int*)hash;
+
+      bloom_byte = *bbp & 0xfffff;
       bloom_bit = 1 << (bloom_byte >> 15);
 
       if (bloom[(bloom_byte & 0x7fff)] & bloom_bit) {
@@ -1279,7 +1313,9 @@ ophtask_t *ophcrack_bforce_lm(ophcrack_t *crack, ophtask_t *task, int id) {
 
       lmtable_mkhash(pwd, hash);
 
-      bloom_byte = (* (int*)hash) & 0xfffff;
+      int *bbp = (int*)hash;
+
+      bloom_byte = *bbp & 0xfffff;
       bloom_bit = 1 << (bloom_byte >> 15);
 
       if (bloom[(bloom_byte & 0x7fff)] & bloom_bit) {
@@ -1692,10 +1728,10 @@ void ophcrack_preload_check(ophcrack_t *crack) {
   while (size--) {
     table_t *tbl = list_rem_head(active);
 
-    if (tbl->preload == 3)
+    if (tbl->idxmem && tbl->endmem && tbl->srtmem)
       list_add_tail(active, tbl);
     
-    else if (tbl->preload == 2 && flag) {
+    else if (tbl->idxmem && tbl->endmem && flag) {
       list_add_tail(active, tbl);
       flag = 0;
     }
@@ -1728,369 +1764,190 @@ ophtask_t *ophcrack_preload_all(ophcrack_t *crack, ophtask_t *task, int id) {
   list_t *active = crack->active;
   list_t *enabled = crack->enabled;
   list_t *remaining = crack->remaining;
-  list_nd_t *nd = 0;
 
-  uint64_t freeram0 = find_freeram();
+  uint64_t freeram0 = crack->find_freeram();
   uint64_t freeram1 = crack->freeram;
   uint64_t freeram  = MY_MIN(freeram0, freeram1);
 
-  /* Build an array which indicates, for a given table id, the desired
-     preloading status of the corresponding table. */
+  /* Compute the amount of memory available if all tables were
+     unloaded. */
+
+  uint8_t *initialised = (uint8_t*)calloc(crack->maxtkind, sizeof(uint8_t));
+
+  for (list_nd_t *nd = crack->tables->head; nd != NULL; nd = nd->next) {
+    table_t *tbl = nd->data;
+    table_preload_t state = table_preload_state(tbl);
+
+    assert(tbl->kind >= 0 && tbl->kind < crack->maxtkind);
+
+    if (tbl->shared_init && initialised[tbl->kind])
+      freeram += table_preload_size(tbl, state & ~preload_init);
+    else
+      freeram += table_preload_size(tbl, state);
+
+    initialised[tbl->kind] |= state & preload_init;
+  }
+
+  /* Mark all tables as inactive and assume that nothing as been
+     preloaded yet. */
+
+  table_preload_t *preload = (table_preload_t*)calloc(crack->maxtid, sizeof(table_preload_t));
   
-  int maxtid = crack->maxtid;
-  int *preload = (int*)calloc(maxtid, sizeof(int));
-
-  /* Mark all the tables as inactive and unload all the tables which
-     have been preloaded but are now disabled. */
-
-  for (nd = crack->tables->head; nd != 0; nd = nd->next) {
+  for (list_nd_t *nd = crack->tables->head; nd != NULL; nd = nd->next) {
     table_t *tbl = nd->data;
 
-    tbl->active = 0;  
-    preload[tbl->id] = tbl->preload;
+    assert(tbl->id >= 0 && tbl->id < crack->maxtid);
 
-    if (!tbl->enabled && tbl->preload > 0) {
-      ophtask_t *task = ophtask_alloc(unload);
-      ophload_t *load = task->data;
-      
-      load->tbl   = tbl;
-      load->count = tbl->preload;
-      
-      freeram += table_preload_size(tbl);
-      scheduler_add(sched, task, disk);
-      
-      preload[tbl->id] = 0;
-    }
+    preload[tbl->id] = preload_none;
+    initialised[tbl->kind] = 0;
+
+    tbl->active = 0;
   }
 
-  /* Build a list which contains all the tables which have been
-     preloaded and are not used any more, either because they have
-     been fully searched in or because that was explicitly
-     requested. */
-
-  list_t *inactive = list_alloc();
-
-  /* If no tables were specified, then we add all the currently
-     preloaded active tables to the list of inactive tables. */
+  /* If tables are provided in the task, we process them. Otherwise we
+     process the active and remaining tables which are enabled.*/
 
   if (tables->size == 0) {
-    for (nd = active->head; nd != 0; nd = nd->next) { 
+    for (list_nd_t *nd = active->head; nd != NULL; nd = nd->next) {
       table_t *tbl = nd->data;
 
-      if (preload[tbl->id] > 0)
-	list_add_tail(inactive, tbl);
+      if (tbl->enabled) 
+	list_add_tail(tables, tbl);
     }
 
-    for (nd = remaining->head; nd != 0; nd = nd->next) {
+    for (list_nd_t *nd = remaining->head; nd != NULL; nd = nd->next) {
       table_t *tbl = nd->data;
-      list_add_tail(tables, tbl);
-    }
-  }
-
-  /* Othewise, we mark as inactive all the tables which have been
-     preloaded and fully covered. */
-
-  else {
-    int n = tables->size;
-    int i;
-
-    for (i=0; i<n; ++i) {
-      table_t *tbl = list_rem_head(tables);
-
-      assert(tbl->enabled);
-
-      if (tbl->cmax == 0) {
-	if (preload[tbl->id] > 0)
-	  list_add_tail(inactive, tbl);
-      } else
+      
+      if (tbl->enabled) 
 	list_add_tail(tables, tbl);
     }
   }
 
-  /* Preload the requested tables. */
-
-#ifndef NDEBUG
-  int ntables = tables->size;
-#endif
+  /* Try to preload as much as we can without preloading .start and
+     with at most one .bin on disk. */
 
   list_clear(active);
   list_clear(remaining);
 
+  table_preload_t states[3] = {
+    preload_init + preload_idx + preload_end,
+    preload_init + preload_idx,
+    preload_init
+  };
 
-  /* If we force the preload, we unload all the tables in use and restart
-     from scratch. */
+  int ondisk = 0;
 
-  if (force_preload < 3) 
-    while (tables->size > 0) {
-      table_t *tbl = list_rem_head(tables);
-      if (preload[tbl->id] > 0) freeram += tbl->idxsize;
-      if (preload[tbl->id] > 1) freeram += tbl->endsize;
-      if (preload[tbl->id] > 2) freeram += tbl->srtsize;
-      
-      preload[tbl->id] = 0;
-      list_add_tail(remaining, tbl);
-    }
+  for (list_nd_t *nd = tables->head; nd != NULL; nd = nd->next) {
+    table_t *tbl = nd->data;
 
-  while (tables->size > 0) {
-    table_t *tbl = list_rem_head(tables);
-    uint64_t size = 0;
+    /* Tables fully covered are never preloaded */
 
-    assert(tbl->enabled);
+    if (tbl->cmax == 0) continue;
 
-    if (preload[tbl->id] < 1) size += tbl->idxsize;
-    if (preload[tbl->id] < 2) size += tbl->endsize;
-    if (preload[tbl->id] < 3) size += tbl->srtsize;
+    for (int i=0; i<3; ++i) {
+      assert(tbl->enabled);
 
-    /* Check if we can preload all the files. */
+      /* Compute the amount of memory required by the preload */
 
-    if (size <= freeram) {
-      freeram -= size;
-
-      preload[tbl->id] = 3;
-      list_add_tail(active, tbl);
-    }
-
-    /* Otherwise, we try to unload some or all of the previously
-       loaded tables, until we have enough space to perform the
-       preload. */
-
-    else {
-      /* We first try to get additional space by unloading some files
-	 of the inactive tables. */
-
-      while (size > freeram && inactive->size > 0) {
-	table_t *tbl = inactive->head->data;
-	uint64_t size = 0;
-
-	assert(preload[tbl->id] > 0);
-
-	if (preload[tbl->id] == 3) size += tbl->srtsize;
-	else if (preload[tbl->id] == 2) size += tbl->endsize;
-	else if (preload[tbl->id] == 1) size += tbl->idxsize;
-	
-	freeram += size;
-	--preload[tbl->id];
-
-	if (preload[tbl->id] == 0)
-	  list_rem_head(inactive);
-      }
-
-      assert(size <= freeram || (size > freeram && inactive->size == 0));
-
-      /* If there is still not enough space, then we try to unload as
-	 many files as needed from the preloaded and active tables. */
-
-      while (size > freeram && tables->size > 0) {
-	table_t *tbl = tables->tail->data;
-
-	/* If some files have been preloaded, then we unload them one
-	   after another. */
-
-	if (preload[tbl->id] > 0) {
-	  uint64_t size = 0;
-
-	  if (preload[tbl->id] == 3) size += tbl->srtsize;
-	  else if (preload[tbl->id] == 2) size += tbl->endsize;
-	  else if (preload[tbl->id] == 1) size += tbl->idxsize;
-
-	  freeram += size;
-	  --preload[tbl->id];
-	}
-
-	/* Otherwise, we add the table to the list of tables which we
-	   will try to preload next time. */
-
-	else {
-	  assert(preload[tbl->id] == 0);
-
-	  list_rem_tail(tables);
-	  list_add_head(remaining, tbl);
-	}
-      }
-
-      assert(size <= freeram || (size > freeram && tables->size == 0));
-
-      /* At this point, either there is enough space or all the
-	 previously preloaded and active tables have been unloaded. */
-      
-      if (size <= freeram ) {
-	freeram -= size;
-	preload[tbl->id] = 3;
-	list_add_tail(active, tbl);
-      } 
-      
-      /* If there is not enough space, then we add the table to the
-	 list of tables which we will try to preload next time. */
-      
-      else {
-	assert(tables->size == 0);
-
-	if (preload[tbl->id] > 0) freeram += tbl->idxsize;
-	if (preload[tbl->id] > 1) freeram += tbl->endsize;
-	if (preload[tbl->id] > 2) freeram += tbl->srtsize;
-
-	preload[tbl->id] = 0;
-	list_add_head(remaining, tbl);
-      }
-    }
-  }
-
-  assert(tables->size == 0);
-
-  /* See if some tables in the 'remaining' list can be completely
-     preloaded and if at most one table can be partially preloaded. */
-
-  int n = remaining->size;
-  int i, flag = 1;
-
-  for (i=0; i<n; ++i) {
-    table_t *tbl = list_rem_head(remaining);
-    uint64_t size = 0;
-
-    assert(preload[tbl->id] == 0);
-
-    size += tbl->idxsize;
-    size += tbl->endsize;
-    size += tbl->srtsize;
-
-    /* Check if we can preload all the files. */
-
-    if (size <= freeram && force_preload == 3) {
-      freeram -= size;
-      
-      preload[tbl->id] = 3;
-      list_add_tail(active, tbl);
-    } 
-  
-    /* Check if we can preload at least the .index and .bin files. If
-       we can, then we set the flag to zero, since we want at most one
-       table partially preloaded. */
-
-    else if (flag) {
-      size -= tbl->srtsize;
-      
-      if (size <= freeram && force_preload > 1) {
-	freeram -= size;
-	
-	preload[tbl->id] = 2;
-	list_add_tail(active, tbl);
-      } else
-	list_add_tail(remaining, tbl);
-
-      /* If we want to force the preload of less files, we do not change
-	 the flag in order to load more than one table if possible. */
-
-      if (force_preload == 3) flag = 0;
-    }
-
-    /* If the table cannot be preloaded either completely or
-       partially, then we put it back into the 'remaining' list. */
-
-    else
-      list_add_tail(remaining, tbl);
-  }
-
-  /* If no tables could be preloaded, then we try to preload at least
-     the .index file of the first remaining table. */
-  
-  if (active->size == 0 && remaining->size > 0 ) {
-    if (force_preload > 1) {
-      table_t *tbl = remaining->head->data;
+      table_preload_t state = states[i];
       uint64_t size = 0;
-      
-      assert(preload[tbl->id] == 0);
-      size += tbl->idxsize;
-      
-      if (size <= freeram && force_preload > 0) {
-	freeram -= size;
-	
-	preload[tbl->id] = 1;
-	list_rem_head(remaining);
-	list_add_tail(active, tbl);
-      }
-    } else {
-      
-      /* If we force to preload only the index files, then we want to
-	 preload all the indexes we can. */
-      
-      n = remaining->size;
-      
-      for (i=0; i<n; ++i) {
-	table_t *tbl = list_rem_head(remaining);
-	uint64_t size = 0;
-	
-	assert(preload[tbl->id] == 0);
 
-	size += tbl->idxsize;
-	
-	if (size <= freeram && force_preload > 0) {
-	  freeram -= size;
-	  
-	  preload[tbl->id] = 1;
-	  list_add_tail(active, tbl);
-	} else
-	  list_add_tail(remaining, tbl);
+      if (tbl->shared_init && initialised[tbl->kind])
+	size = table_preload_size(tbl, state & ~preload_init);
+      else
+	size = table_preload_size(tbl, state);
+
+      /* If one .idx or one .end is already on disk, only accept to
+	 preload tables for which .idx and .end can both be
+	 preloaded. */
+
+      static const table_preload_t idxend = preload_idx + preload_end;
+      int accept = !ondisk || (ondisk && (state & idxend) == idxend);
+
+      if (accept && size <= freeram) {
+	freeram -= size;
+	preload[tbl->id] = state;
+	initialised[tbl->kind] |= state & preload_init;
+	break;
+      }
+    }
+
+    ondisk = ondisk || (!(preload[tbl->id] & preload_idx) ||
+			!(preload[tbl->id] & preload_end));
+
+    if (preload[tbl->id] == preload_none)
+      list_add_tail(remaining, tbl);
+    else
+      list_add_tail(active, tbl);
+  }
+
+  /* Try to preload .start for tables for which .index and .bin as
+     already been preloaded. */
+
+  for (list_nd_t *nd = active->head; nd != NULL; nd = nd->next) {
+    table_t *tbl = nd->data;
+
+    if (preload[tbl->id] & (preload_idx + preload_end)) {
+      uint64_t size = table_preload_size(tbl, preload_srt);
+
+      if (size <= freeram) {
+	freeram -= size;
+	preload[tbl->id] += preload_srt;
       }
     }
   }
   
-  /* We mark all the newly preloaded tables as active. */
+  /* Mark tables in the active list as active and tables in the
+     remaining list as inactive. */
   
-  for (nd = active->head; nd != 0; nd = nd->next) {
+  for (list_nd_t *nd = active->head; nd != NULL; nd = nd->next) {
     table_t *tbl = nd->data;
     tbl->active = 1;
   }
 
-  /* We mark all the remaining tables as inactive if at least one
-     table has been preloaded. We mark only the first one as active
-     otherwise. 
+  for (list_nd_t *nd = remaining->head; nd != NULL; nd = nd->next) {
+    table_t *tbl = nd->data;
+    tbl->active = 0;
+  }
 
-     If we force not to preload, then we mark all tables as active. */
-  
-  if (force_preload > 0) {
-    for (nd = remaining->head; nd != 0; nd = nd->next) {
-      table_t *tbl = nd->data;
-      tbl->active = 0;
-    }
-   
-    if (active->size == 0 && remaining->size > 0) {
-      table_t *tbl = list_rem_head(remaining);
-      
-      tbl->active = 1;
-      list_add_tail(active, tbl);
-      
-      assert(active->size == 1);
-    } 
-  } else {
-    while (remaining->size > 0) {
-      table_t *tbl = list_rem_head(remaining);
-      tbl->active = 1;
-      list_add_tail(active, tbl);
+  /* Schedule an unload task for all tables which have been at least
+     partially preloaded and have been disabled. */
+
+  for (list_nd_t *nd = crack->tables->head; nd != 0; nd = nd->next) {
+    table_t *tbl = nd->data;
+    table_preload_t state = table_preload_state(tbl);
+
+    if (!tbl->enabled && state != preload_none) {
+      ophtask_t *task = ophtask_alloc(unload);
+      ophload_t *load = task->data;
+
+      load->tbl = tbl;
+      load->preload = state;
+
+      scheduler_add(sched, task, disk);
     }
   }
 
-  list_free(inactive);
-  assert(active->size + remaining->size == ntables);
-
-  /* Scheduler the un/preload tasks only is there is actually
+  /* Schedule the un/preload tasks only if there is actually
      something to do with the tables currently activated. */
 
   uint64_t total_size = 0;
 
   if (ophcrack_has_next(crack)) {
-    /* Schedule an unload task for all the tables for which the desired
-       status is lower than the current one. */
+    /* Schedule an unload task for all the tables for which the
+       desired preload status is different from the current one. */
     
-    for (nd = enabled->head; nd != 0; nd = nd->next) {
+    for (list_nd_t *nd = enabled->head; nd != 0; nd = nd->next) {
       table_t *tbl = nd->data;
-      int n = tbl->preload - preload[tbl->id];
-      
-      if (n > 0) {
+      table_preload_t state = table_preload_state(tbl);
+      table_preload_t unload_bits = state & ~preload[tbl->id];
+
+      if (unload_bits != preload_none) {
 	ophtask_t *task = ophtask_alloc(unload);
 	ophload_t *load = task->data;
 	
 	load->tbl = tbl;
-	load->count = tbl->preload - preload[tbl->id];
+	load->preload = unload_bits;
 	
 	scheduler_add(sched, task, disk);
       }
@@ -2104,32 +1961,49 @@ ophtask_t *ophcrack_preload_all(ophcrack_t *crack, ophtask_t *task, int id) {
     scheduler_add(sched, unload_done, disk);
     
     /* Schedule a preload task for all the active tables for which the
-       desired preload status is higher than the current one. */
-    
-    for (nd = active->head; nd != 0; nd = nd->next) {
+       desired preload status is different from the current one. */
+
+    for (list_nd_t *nd = enabled->head; nd != 0; nd = nd->next) {
       table_t *tbl = nd->data;
+      table_preload_t state = table_preload_state(tbl);
+      table_preload_t load_bits = ~state & preload[tbl->id];
       
-      if (preload[tbl->id] > tbl->preload) {
-	int n = preload[tbl->id];
-	int i;
+      if (load_bits == preload_none)
+	continue;
+
+      total_size += table_preload_size(tbl, load_bits);
+
+      static const table_preload_t steps[4] = { 
+	preload_init, 
+	preload_idx, 
+	preload_end, 
+	preload_srt 
+      };
 	
-	for (i=tbl->preload; i<n; ++i) {
+      for (int i=0; i<4; ++i) {
+	if (load_bits & steps[i]) {
 	  ophtask_t *task = ophtask_alloc(preload_one);
 	  ophload_t *load = task->data;
-	  uint64_t size = 0;
-	  
-	  if (i == 0) size = tbl->idxsize;
-	  else if (i == 1) size = tbl->endsize;
-	  else if (i == 2) size = tbl->srtsize; 
-	  
-	  total_size += size;
-	  
+	    
 	  load->tbl = tbl;
+	  load->preload = steps[i];
+
 	  scheduler_add(sched, task, disk);
 	}
       }
     }
   }
+
+  /* If no tables have been activated and some tables remain to be
+     preloaded, it means the remaining tables have to be initialised,
+     but there is not enough memory available. */
+
+  if (active->size == 0) {
+    while (remaining->size > 0) {
+      table_t *tbl = list_rem_head(remaining);
+      assert(tbl->inisize > 0);
+    }
+  }
   
   /* Send a message indicating the number of tables preloaded and the
      amount of RAM they will occupy. */
@@ -2144,6 +2018,7 @@ ophtask_t *ophcrack_preload_all(ophcrack_t *crack, ophtask_t *task, int id) {
   scheduler_add(sched, preload_done, disk);
 
   free(preload);
+  free(initialised);
 
   return task;
 }
@@ -2158,22 +2033,18 @@ ophtask_t *ophcrack_preload_one(ophcrack_t *crack, ophtask_t *task, int id) {
   if (tbl != 0) {
     assert(tbl->enabled);
 
-    uint64_t size = 0;
+    uint64_t size = table_preload_size(tbl, load->preload);
 
-    if (tbl->preload == 0)      size = tbl->idxsize;
-    else if (tbl->preload == 1) size = tbl->endsize;
-    else if (tbl->preload == 2) size = tbl->srtsize;
-
-    message_preload(tbl, tbl->preload+1, 0, size);
-    size = table_preload(tbl);
+    message_preload(tbl, load->preload, 0, size);
+    size = table_preload(tbl, load->preload);
 
     pthread_mutex_lock(mutex);
     crack->freeram -= size;
     pthread_mutex_unlock(mutex);
 
-    message_preload(tbl, tbl->preload, 1, size);    
-  } 
-
+    table_preload_t state = table_preload_state(tbl);
+    message_preload(tbl, state, 1, size);
+  }
   else
     message_done(preload_all);
 
@@ -2186,29 +2057,16 @@ ophtask_t *ophcrack_unload(ophcrack_t *crack, ophtask_t *task, int id) {
 
   ophload_t *load = task->data;
   table_t *tbl = load->tbl;
-  
+
   if (tbl != 0) {
-    int count = load->count;
-    assert(tbl->preload >= count);
+    uint64_t size = table_unload(tbl, load->preload);
 
-    uint64_t size = 0;
-    int i;
-  
-    for (i=0; i<count; ++i) {
-      if (tbl->preload == 3)      size += tbl->srtsize;
-      else if (tbl->preload == 2) size += tbl->endsize;
-      else if (tbl->preload == 1) size += tbl->idxsize;
+    pthread_mutex_lock(mutex);
+    crack->freeram += size;
+    pthread_mutex_unlock(mutex);
 
-      size = table_unload(tbl);
-
-      pthread_mutex_lock(mutex);
-      crack->freeram += size;
-      pthread_mutex_unlock(mutex);
-
-      message_unload(tbl, size);
-    }
+    message_unload(tbl, size);
   }
-
   else
     message_done(unload);
 
